/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// BaseCmds.cp -- basic commands

#include <stdlib.h>
#include "mh2html.h"
#include "files.h"
#include "charhdl.h"
#include "basecmds.h"
#include "baseenvs.h"
#include "counter.h"
#include "utils.h"

static int gBreakLevel = 6;

// set level where to break current output file and start new one

void
SetBreakLevel(int lev) {
    gBreakLevel = lev;
}

// \verb... command

VerbCmd gVerbCmd;

bool
VerbCmd::translate(Stream *s) {
    char c, e = s->get();
    Puts("<code>");
    while ((c = s->get()) != e) {
        if (c == kEndChar) {
            s->fatal("run-away \verb argument");
            return false;
        }
        PutVerbc(c);
    }
    Puts("</code>");
    return true;
}

bool
VerbCmd::scan(Stream *s) {
    char c, e = s->get();
    while ((c = s->get()) != e) {
        if (c == kEndChar) return false;
    }
    return true;
}

// section commands which may break HTML file into parts

void
BreakingCmd::nextFile(Stream *s) {
    if (fLevel <= gBreakLevel) {
      //	PutFooter(s);
	OpenNextHTMLFile(s);
	//	PutHeader(s);
    }
}

void
BreakingCmd::putContents(Stream *s, const char *title) {
    // eject page
    nextFile(s);
    // printe header into contents file
    if (OpenNewContFile()) {
        PutConts("\\tocpageline{");
	PutConts(fLevelStr);
	PutConts("}{");
	PutConts(GetHTMLFileName());
	PutConts("}{");
        PutConts(title);
        PutConts("}{}\n");
    } else s->error("could not open contents file");
}

class SectionCmd : public CmdHdl, public BreakingCmd {
public:
    SectionCmd(int lev) : CmdHdl(1, 1), BreakingCmd(lev) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
};

bool
SectionCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    putContents(s, a[0]->cont());
    // increase counter
    if(0 < fLevel && fLevel <= kNumSectCounter)
      ++gSectCounter[fLevel-1];
    // put header
    Puts("<h");
    Puts(fLevelStr);
    Putc('>');
    /*
    if(fLevel <= kNumSectCounter) {
      char buff[10];
      sprintf(buff, "%d ", gSectCounter[fLevel-1].get());
      Puts(buff);
    }
    */
    gScanner->donePar();
    gScanner->translate(a[0]);
    Puts("</h");
    Puts(fLevelStr);
    Putc('>');
    gScanner->markPar();
    return true;
}

Counter *gSectCounter;
static SectionCmd chapterCmd(1);
static SectionCmd sectionCmd(2);
static SectionCmd subsectionCmd(3);
static SectionCmd subsubsectionCmd(4);
static SectionCmd paragraphCmd(5);

// input read additional files

static class InputCmd : public CmdHdl {
public:
    InputCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} inputCmd;

bool
InputCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    FILE *inf;
    char buff[256];
    if (! a[0]->copy(buff, 252)) {
        a[0]->error("path name too long, \\input ignored");
        return true;
    }
    int n = strlen(buff);
    if (n <= 4 || strcmp(buff+n-4, ".tex") != 0)
        strcat(buff, ".tex");
    if ((inf = fopen(buff, READ_MODE)) == NULL) {
        a[0]->error("could not open \"",buff,"\", \\input ignored");
        return true;
    }
    FileStream inStream(inf, buff);
    Scanner *sc = gScanner;
    gScanner = &gTextScanner;
    gScanner->translate(&inStream);
    gScanner = sc;
    return true;
}

// title command

// the default handler puts a \title{#1} line into the *.hhd file

static class TitleCmd : public CmdHdl {
public:
    TitleCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} titleCmd;

bool
TitleCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    // printer title into header file
    if (OpenNewHeaderFile()) {
        PutHeaders("\\title{");
        PutHeaders(a[0]->cont());
        PutHeaders("}\n");
    } else {
        s->error("could not open header file");
        return false;
    }
    return true;
}

// the following handler is active when reading the *.hco file

static class TocpagelineCmd : public CmdHdl {
protected:
    int fLevel;
public:
    void initLevel() { fLevel = 0; }
    int getLevel() { return fLevel; }
    TocpagelineCmd() : CmdHdl(0, 4), fLevel(0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} tocpagelineCmd;

bool
TocpagelineCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    int l = atoi(a[0]->cont());
    const char *c;
    int indent;
    if (l == 0) {
	s->error("error in contents file [illegal level]");
	return true;
    }
    if (l > fLevel) {
	for (; fLevel < l; fLevel++) {
          for(indent=0; indent < fLevel; ++indent)
            Puts("    ");
          Puts("<ol>\n");
        }
    } else if (l < fLevel) {
	for (; fLevel > l; fLevel--) {
          for(indent=0; indent < fLevel; ++indent)
            Puts("    ");
          Puts("</ol>\n");
        }
    }
    if ( strncmp ( a[1]->cont(), "./", 2 ) == 0 )
      c = a[1]->cont() + 2;
    else
      c = a[1]->cont();
    gFilenames.push_back(c);
    for(indent=0; indent < fLevel; ++indent)
      Puts("    ");
    Puts("<li><a href=\"");
    Puts(a[1]->cont());
    Puts("\">");
    gScanner->translate(a[2]);
    a[2]->reset();
    Puts("</a>");
    if(strlen(a[3]->cont()) > 0) {
      Puts(" - ");
      // redefine MInfo environment
      gEnvs.push("MInfo", &gNullEnv);
      gScanner->translate(a[3]);
      a[3]->reset();
      gEnvs.pop("MInfo");
    }
    //   Putnl();
    return true;
}

// the following handler is active when reading the *.hco file

static class TitleContCmd : public CmdHdl {
public:
    TitleContCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} titleContCmd;

bool
TitleContCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    Puts("<h1>");
    gScanner->translate(a[0]);
    Puts("</h1>\n");
    return true;
}

// put contents for first page

void
PutContents(Stream *s) {
    FILE *oldHeaderf, *oldContf;

    // put title from *.hhd file
    if ( (oldHeaderf = OpenOldHeaderFile()) ) {
        // redefine \title locally
        gCmds.push("title", &titleContCmd);
        FileStream inStream(oldHeaderf, ".hhd");
        Scanner *sc = gScanner;
        gScanner = &gTextScanner;
        gScanner->translate(&inStream);
        gScanner = sc;
        gCmds.pop("title");
    } else {
        s->warning("no header file present, run again to fix contents page");
    }

    // put contents from *.hco file
    if ( (oldContf = OpenOldContFile()) ) {
	Puts("<p>\n");
	gScanner->donePar();
        // redefine \tocpageline locally
	tocpagelineCmd.initLevel();
        gCmds.push("tocpageline", &tocpagelineCmd);
        FileStream inStream(oldContf, ".hco");
        Scanner *sc = gScanner;
        gScanner = &gTextScanner;
        localParaType _(kNone);
        gScanner->translate(&inStream);
        gScanner = sc;
        gCmds.pop("tocpageline");

	// close lists opened with tocpagelineCmd::translate
	int l = tocpagelineCmd.getLevel();
	for (; l > 0; l--) Puts("</ol>\n");
	Puts("</p>\n");
    } else {
        s->warning("no contents file present, run again to fix contents page");
    }
}

// Lables and references
std::map<std::pair<std::string,std::string>, int> gLocalLables;
std::map<std::string,int> gLables;

static class MakeLocalLabelCmd : public CmdHdl {
public:
  MakeLocalLabelCmd() : CmdHdl(0, 3) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} makelocallabelCmd;

bool
MakeLocalLabelCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
  int num;
  sscanf(a[2]->cont(), "%d", &num);
  gLocalLables[std::pair<std::string,std::string>(a[0]->cont(),a[1]->cont())]
    = num;
  return true;
}

static class MakeLabelCmd : public CmdHdl {
public:
  MakeLabelCmd() : CmdHdl(0, 2) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} makelabelCmd;

bool
MakeLabelCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
  int num;
  sscanf(a[1]->cont(), "%d", &num);
  gLables[string(a[0]->cont())]
    = num;
  return true;
}

static class RefCmd : public CmdHdl {
public:
  RefCmd() : CmdHdl(0, 1) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} refCmd;

bool
RefCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
  std::pair<std::string,std::string> target(GetHTMLFileName(), a[0]->cont());
  
  Puts("<a href=\"");
  //  Puts(gDefs.top("thisfile"));
  Putc('#');
  Puts(a[0]->cont());
  Puts("\">");
  if ( gLocalLables.find(target) != gLocalLables.end() ) {
    char buff[128];
    sprintf(buff, "%d", (*(gLocalLables.find(target))).second);
    Puts(buff);
  } else
    Puts("[>>]");
  Puts("</a>");
  return true;
}
      

// style-changing commands \em ,\tt, \bf, \it, \sc
// note that only these commands may change the current font style such
// that the change is removed at the end of a {...}-context

class StyleCmd : public CmdHdl {
protected:
    char *style;
    void (*rmStyle)();
public:
    StyleCmd(char *s, void (*f)()) : CmdHdl(0, 0) { 
        style = s; rmStyle = f; 
    }
    bool translate(Stream *);
};

bool
StyleCmd::translate(Stream *s) {
    Putc('<'); Puts(style); Putc('>');
    gScanner->atEnd(rmStyle);
    return true;
}

static void rmEm() { Puts("</em>"); }

static class EmCmd : public StyleCmd {
public:
    EmCmd() : StyleCmd("em", rmEm) {};
} emCmd;

static void rmBf() { Puts("</b>"); }

static class BfCmd : public StyleCmd {
public:
    BfCmd() : StyleCmd("b", rmBf) {};
} bfCmd;

static void rmIt() { Puts("</i>"); }

static class ItCmd : public StyleCmd {
public:
    ItCmd() : StyleCmd("i", rmIt) {};
} itCmd;

static void rmTt() { Puts("</tt>"); }

static class TtCmd : public StyleCmd {
public:
    TtCmd() : StyleCmd("tt", rmTt) {};
} ttCmd;

static void rmSc() { Puts("</span>"); }

static class ScCmd : public StyleCmd {
public:
  ScCmd() : StyleCmd("span style=\"font-variant: small-caps;\"", rmSc) {};
} scCmd;

// init the command handlers

void
initBasicCmds() {
    gCmds.push("verb", &gVerbCmd);
    gCmds.push("input", &inputCmd);
    gCmds.push("include", &inputCmd);
    gCmds.push("em", &emCmd);
    gCmds.push("it", &itCmd);
    gCmds.push("sl", &itCmd);
    gCmds.push("bf", &bfCmd);
    gCmds.push("tt", &ttCmd);
    gCmds.push("sc", &scCmd);
    gCmds.push("title", &titleCmd);
    gCmds.push("chapter", &chapterCmd);
    gCmds.push("chapter*", &chapterCmd);
    gCmds.push("section", &sectionCmd);
    gCmds.push("section*", &sectionCmd);
    gCmds.push("subsection", &subsectionCmd);
    gCmds.push("subsection*", &subsectionCmd);
    gCmds.push("subsubsection", &subsubsectionCmd);
    gCmds.push("subsubsection*", &subsubsectionCmd);
    gCmds.push("paragraph", &paragraphCmd);
    gCmds.push("paragraph*", &paragraphCmd);
    gCmds.push("makelabel", &makelabelCmd);
    gCmds.push("makelocallabel", &makelocallabelCmd);
    gCmds.push("ref", &refCmd);
    
    gSectCounter = new Counter[kNumSectCounter];
    for(int i=1; i<kNumSectCounter; ++i)
      gSectCounter[i-1].addChild(&gSectCounter[i]);
}

// end of file
