/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// BaseEnvs.cp -- basic environments

#include <ctype.h>
#include "files.h"
#include "charhdl.h"
#include "basecmds.h"
#include "baseenvs.h"

// globals

static VerbatimEnv verbatimEnv(0, 0, "\\end{verbatim}");
EnvHdl gNullEnv(0, 0);

// the verbatim env

bool 
VerbatimEnv::translateBody(Stream *s) {
    char c;
    const char *em = endMark, *cp;
    Puts("<pre>");
    while (isspace(c = s->get()) && '\n'!=c) /* nix */;
    if('\n'==c) c=s->get();
    for(;;) {
        if (c == kEndChar)
            s->fatal("run-away verbatim environment");
	if (c == *em) {
	    em++;
	    if (*em == '\0') break;
	} else {
	    for (cp = endMark; cp != em; cp++) PutVerbc(*cp);
	    em = endMark;
	    if (c == *em) em++; 
	    else PutVerbc(c);
	}
        c = s->get();
    }
    Puts("</pre>");
    gCurrEnvName.pop();
    return true;
}

bool 
VerbatimEnv::scanBody(Stream *s) {
    char c;
    const char *em = endMark;
    for(;;) {
        c = s->get();
        if (c == kEndChar) return false;
	if (c == *em) {
	    em++;
	    if (*em == '\0') break;
	} else {
	    em = endMark;
	    if (c == *em) em++;
	}
    }
    gCurrEnvName.pop();
    return true;
}

// ignored environments
// xxx MChanges should not be ignored.

static class IgnoredEnv : public VerbatimEnv {
public:
  IgnoredEnv(int o, int a, char *e) : VerbatimEnv(o,a,e) {;}
  virtual bool translateBody(Stream *s) { return(scanBody(s)); }
} NoHTMLEnv(0,0,"\\end{Mnohtml}"), ToDoEnv(0,0,"\\end{todo}"),
  ChangesEnv(0,0,"\\end{MChanges}");

// the tabular env

// & and \\ need special attention:

static class TableFieldSepChar : public CharHdl {
public:
    virtual bool translate(Stream *, char);
} tableFieldSepChar;

bool
TableFieldSepChar::translate(Stream *s, char c) {
    Puts("<td valign=top>\n");
    return true;
}

static class TableLineSepHdl : public CmdHdl {
public:
    TableLineSepHdl() : CmdHdl(1, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} tableLineSepHdl;

bool
TableLineSepHdl::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    Puts("<tr>\n<td valign=top>\n");
    return true;
}

static class TabularEnv : public EnvHdl {
protected:
    CharHdl *fSepHdl;
public:
    TabularEnv() : EnvHdl(1, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
} tabularEnv;

bool
TabularEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[])
{
    // currently the alignment and placement is totally ignored
    EnvHdl::translateArgs(s, oa, no, a);
    fSepHdl = gScanner->getHdl('&');
    // treat & and \\ specially
    gScanner->setHdl('&', &tableFieldSepChar);
    gCmds.push("\\", &tableLineSepHdl);
    Puts("<table border=0>\n<tr>\n<td valign=top>\n");
    gScanner->donePar();
    return true;
}

bool
TabularEnv::translateEnd(Stream *s, SubStream *oa[], int no, SubStream *a[])
{
    // restore default handling of & and backslash
    gScanner->setHdl('&', fSepHdl);
    gCmds.pop("\\");
    EnvHdl::translateEnd(s, oa, no, a);
    Puts("</table>\n");
    return true;
}

// thebibliography environment

static class ThebibliographyEnv : public ParsedEnv, public BreakingCmd {
public:
    ThebibliographyEnv() :
	ParsedEnv(0, 1,
	    "<h2>#(bibliography):</h2>\n<dl>", NULL, "</dl>"),
	BreakingCmd(2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} thebibliographyEnv;

bool
ThebibliographyEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    const char *c;
    c = gDefs.top("bibliography");
    if (! c) c = "Bibliography";
    putContents(s, c);
    return ParsedEnv::translateArgs(s, oa, no, a);
}

// appendix environment

static class AppendixEnv : public ParsedEnv, public BreakingCmd {
public:
    AppendixEnv() :
	ParsedEnv(0, 0,
	    "<h2>#(appendix):</h2>\n<dl>", NULL, "</dl>"),
	BreakingCmd(2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} appendixEnv;

bool
AppendixEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    const char *c;
    c = gDefs.top("appendix");
    if (! c) c = "Appendix";
    putContents(s, c);
    return ParsedEnv::translateArgs(s, oa, no, a);
}

// init the environment handlers

void
initBasicEnvs() {
    gEnvs.push("verbatim", &verbatimEnv);
    gEnvs.push("tabular", &tabularEnv);
    gEnvs.push("longtable", &tabularEnv);
    gEnvs.push("thebibliography", &thebibliographyEnv);
    gEnvs.push("appendix", &appendixEnv);
    gEnvs.push("Mnohtml",&NoHTMLEnv);
    gEnvs.push("todo",&ToDoEnv);
}

// end of file
