/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// CharHdl.cp -- character handlers and descendants

#include <ctype.h>
#include <assert.h>
#include "error.h"
#include "files.h"
#include "stream.h"
#include "charhdl.h"
#include "cmdhdl.h"

// constants

const char kaUml = '';
const char koUml = '';
const char kuUml = '';
const char kAUml = '';
const char kOUml = '';
const char kUUml = '';
const char kSZLig = '';
const char keAcute = '';

// globals

CharHdl gCharHdl;
CmdChar gCmdChar;
IllegalChar gIllegalChar;
IgnoreChar gIgnoreChar;
EndChar gEndChar;
GermanQuoteChar gGermanQuoteChar;
UmlautChar gUmlChar;
SubscriptChar gSubscriptChar;
SuperscriptChar gSuperscriptChar;

TextScanner gTextScanner;
MathScanner gMathScanner;
PlainScanner gPlainScanner;
MupScanner gMupScanner;

Scanner *gScanner = &gTextScanner;

static inline int
is_alpha(int c)
{
    return isascii(c) && isalpha(c);
}

// the base class simply passes any character thru

bool 
CharHdl::translate(Stream *s, char c) {
    Putc(c);
    return true;
}

bool 
CharHdl::scan(Stream *s, char c) {
    return true;
}

void 
CharHdl::par() {
    Puts2("<p>");
    gScanner->donePar();
}

// handle command character '\'

const int kCmdLen = 128;		// max length of command name + 2

SubStream*
CmdChar::getCmd(Stream *s, char cc) {
  std::string str;
  StreamPos p = s->pos();
  char c = s->get();
  str = cc;
  if (!is_alpha(c) || '@' == c) { // single-char command
    str += c;
  } else {
    while(is_alpha(c)) {
      str += c;
      c = s->get();
    }
    // trailing * considered part of command
    if('*' == c)
      str =+ c;
    else
      s->unget(c);
  }
  return new SubStream(s, str, &p);
}

bool
CmdChar::translateCmd(Stream *s, char *c) {
    CmdHdl *h = gCmds.top(c);
    if (h) {
#ifdef DEBUG
	printf("tranlate cmd: %s\n");
#endif
        return h->translate(s);
    }
    else s->error("unknown command '\\",c,"'");
    return true;
}

bool 
CmdChar::translate(Stream *s, char ch) {
    char buff[kCmdLen], *c;
    int i;
    c = buff;
    *c = s->get();
    if (! is_alpha(*c) && *c != '@') {	// one-char command?
        c++;
        *c = '\0';
        return translateCmd(s, buff);
    }
    for (i = 2, c++; i < kCmdLen; i++, c++) {
        *c = s->get();
        if (! is_alpha(*c) && *c != '@') {
            // trailing * ?
            if (*c == '*') c++; else s->unget(*c);
            *c = '\0';
            return translateCmd(s, buff);
        }
    }
    s->unget(*c);
    s->error("run-away command name");
    return true;
}

bool
CmdChar::scanCmd(Stream *s, char *c) {
    CmdHdl *h = gCmds.top(c);
    if (h) {
#ifdef DEBUG
	printf("scan cmd: %s\n");
#endif
	return h->scan(s);
    }
    return true;
}

bool 
CmdChar::scan(Stream *s, char ch) {
    char buff[kCmdLen], *c;
    int i;
    c = buff;
    *c = s->get();
    if (! is_alpha(*c) && *c != '@') {	// one-char command?
        c++;
        *c = '\0';
        return scanCmd(s, buff);
    }
    for (i = 2, c++; i < kCmdLen; i++, c++) {
        *c = s->get();
        if (! is_alpha(*c) && *c != '@') {
            // trailing * ?
            if (*c == '*') c++; else s->unget(*c);
            *c = '\0';
            return scanCmd(s, buff);
        }
    }
    s->unget(*c);
    return true;
}

void 
CmdChar::par() {
  Puts2("<p>");
  gScanner->donePar();
}

// signal end of stream

bool 
EndChar::translate(Stream *s, char c) {
    return false;
}

bool 
EndChar::scan(Stream *s, char c) {
    return false;
}

void 
EndChar::par() {
    gScanner->donePar();
}

// handle illegal char

bool 
IllegalChar::translate(Stream *s, char c) {
    char b[2];
    b[0] = c;
    b[1] = '\0';
    s->warning("illegal character '", b, "'");
    return true;
}

// ignore character

bool 
IgnoreChar::translate(Stream *s, char c) {
    return true;
}

void 
IgnoreChar::par() {}

// subscripts

bool
SubscriptChar::translate(Stream *s, char c) {
  return gCmdChar.translateCmd(s,"sb");
}

// superscripts

bool 
SuperscriptChar::translate(Stream *s, char c) {
  return gCmdChar.translateCmd(s,"sp");
}

// handle quote in german environment

bool 
GermanQuoteChar::translate(Stream *s, char c) {
    char buff[10];
    c = s->get();
    switch (c) {
    case '`':
        Puts("&bdquo;");
        break;
    case '\'':
        Puts("&ldquo;");
        break;
    case 's':
        Puts("&szlig;");
        break;
    case 'a':
    case 'A':
    case 'o':
    case 'O':
    case 'u':
    case 'U':
        sprintf(buff, "&%cuml;", c);
        Puts(buff);
        break;
    default:
        s->unget(c);
        Puts("&quot;");
    }
    return true;
}

bool 
GermanQuoteChar::scan(Stream *s, char c) {
    c = s->get();
    switch (c) {
    case '`':
    case '\'':
    case 's':
    case 'a':
    case 'A':
    case 'o':
    case 'O':
    case 'u':
    case 'U':
        break;
    default:
        s->unget(c);
    }
    return true;
}

bool 
UmlautChar::translate(Stream *s, char c) {
    switch (c) {
    case kSZLig:
        Puts("&szlig;");
        break;
    case kaUml:
        Puts("&auml;");
        break;
    case koUml:
        Puts("&ouml;");
        break;
    case kuUml:
        Puts("&uuml;");
        break;
    case kAUml:
        Puts("&Auml;");
        break;
    case kOUml:
        Puts("&Ouml;");
        break;
    case kUUml:
        Puts("&Uuml;");
        break;
    case keAcute:
      Puts("&eacute;");
      break;
    default:
        assert(false);
        Putc(c);
    }
    return true;
}

// handle white space

class WhiteHdl : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
} whiteHdl;

bool
WhiteHdl::translate(Stream *s, char c) {
    Putc2(c);
    int nl = ((c == '\n') ? 1 : 0);
    for (;;) {
        c = s->get();
        if (c == '\n') {
            nl++;
            Putc2(c);
        } else if ((c == ' ') || (c == '\t') || (c == '\r')) {
            Putc2(c);
        } else if (c == kEndChar) {
            if (nl > 1) gScanner->markPar();
            return false;
        } else {
            s->unget(c);
            if (nl > 1) gScanner->markPar();
            return true;
        }
    }
}

// handle formal arguments

class ArgHdl : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
} argHdl;

bool
ArgHdl::translate(Stream *s, char c) {
    s->error("formal argument (#) ignored");
    return true;
}

// handle inline math $...

class MathHdl : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
} mathHdl;

bool
MathHdl::scan(Stream *s, char c) {
  /*    Scanner *sc = gScanner;
	gScanner = &gMathScanner;
    gScanner->scan(s);	// eats the closing $
    gScanner = sc; */
    return true;
}

bool
MathHdl::translate(Stream *s, char c) {
    Puts("<var>");
    Scanner *sc = gScanner;
    gScanner = &gMathScanner;
    gScanner->translate(s); // eats the closing $
    gScanner = sc;
    Puts("</var>");
    return true;
}

// handle space ~

static class FixedSpaceHdl : public CharHdl {
public:
    virtual bool translate(Stream *, char);
} fixedSpaceHdl;

bool
FixedSpaceHdl::translate(Stream *s, char c) {
    Puts("&nbsp;");
    return true;
}

// handle table field seperator & in normal text

class TabOnlyHdl : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
} tabOnlyHdl;

bool
TabOnlyHdl::translate(Stream *s, char c) {
    s->error("table field seperator (&) ignored");
    return true;
}

// handle math chars _ and ^ in normal text

class MathOnlyHdl : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
} mathOnlyHdl;

bool
MathOnlyHdl::translate(Stream *s, char c) {
    Putc(c);
    char b[2] = { c, '\0' };
    s->warning("math command (", b, ") outside of math environment");
    return true;
}

// handle open brace {

class OpenHdl : public CharHdl {
public:
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
} openHdl;

bool
OpenHdl::scan(Stream *s, char c) {
    CharHdl *h = gScanner->getHdl('}');
    gScanner->setHdl('}', &gEndChar);
    gScanner->scan(s);	// eats the closing }
    gScanner->setHdl('}', h);
    return true;
}

bool
OpenHdl::translate(Stream *s, char c) {
    CharHdl *h = gScanner->getHdl('}');
    gScanner->setHdl('}', &gEndChar);
    gScanner->translate(s);	// eats the closing }
    gScanner->setHdl('}', h);
    return true;
}

// handle closing brace } not preceeded by opening brace

class CloseHdl : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
} closeHdl;

bool
CloseHdl::translate(Stream *s, char c) {
    s->error("closing brace } without opening one, ignored");
    return true;
}

// handle comment chars %

class CommentHdl : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
} commentHdl;

bool
CommentHdl::scan(Stream *s, char c) {
    for(;;) {
        c = s->get();
        if (c == '\n') return true;
        if (c == kEndChar) return false;
    }
}

bool
CommentHdl::translate(Stream *s, char c) {
    return scan(s, c);
}

// handle hyphen's

class HyphenHdl : public CharHdl {
public:
    virtual bool translate(Stream *, char);
} hyphenHdl;

bool
HyphenHdl::translate(Stream *s, char c) {
    c = s->get();
    if (c == '-') c = s->get();
    if (c == '-') Putc('-'); else s->unget(c);
    Putc('-');
    return true;
}

// handle HTML-specials <, > and "

static class LtHTMLHdl : public CharHdl {
public:
    virtual bool translate(Stream *, char);
} ltHTMLHdl;

bool
LtHTMLHdl::translate(Stream *s, char c) {
    Puts("&lt;");
    return true;
}

static class GtHTMLHdl : public CharHdl {
public:
    virtual bool translate(Stream *, char);
} gtHTMLHdl;

bool
GtHTMLHdl::translate(Stream *s, char c) {
    Puts("&gt;");
    return true;
}

static class QuotHTMLHdl : public CharHdl {
public:
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
} quotHTMLHdl;

bool
QuotHTMLHdl::translate(Stream *s, char c) {
    char c2;
    c = s->get();
    switch (c) {
    case '`':
        Puts("&bdquo;");
        break;
    case '\'':
        Puts("&ldquo;");
        break;
    case '<':
      Puts("&laquo;");
      break;
    case '>':
      Puts("&raquo;");
      break;
    case 's':
      Puts("&szlig;");
      break;
    case 'a': case 'e': case 'i': case 'o': case 'u':
    case 'A': case 'E': case 'I': case 'O': case 'U':
      Putc('&'); Putc(c); Puts("uml;");
      break;
    case 'f': case 'l': case 'm': case 'p': case 'r': case 't':
    case 'F': case 'L': case 'M': case 'P': case 'R': case 'T':
      c2 = s->get();
      if (c == c2) {
	Putc(c);
	Putc(c);
      } else {
	s->unget(c2);
	s->unget(c);
	Puts("&quot;");
      }
      break;
    case 'c':
      c2 = s->get();
      if ('k' == c2) {
	Putc(c);
	Putc(c2);
      } else {
	s->unget(c2);
	s->unget(c);
	Puts("&quot;");
      }
      break;
    case '-':
    case '|':
    case '"':
      break;
    case '~':
    case '=':
      Putc('-');
      break;
    default:
        s->unget(c);
        Puts("&quot;");
    }
    return true;
}

bool 
QuotHTMLHdl::scan(Stream *s, char c) {
    c = s->get();
    switch (c) {
    case '`':
    case '\'':
        break;
    default:
        s->unget(c);
    }
    return true;
}

// character scanners base class

Scanner::Scanner() {
    int i;
    for (i = 0; i < 256; i++) {
        hdl[i] = &gIllegalChar;
    }
    for (i = 32; i < 127; i++) {
        hdl[i] = &gCharHdl;
    }
    hdl[kaUml] = &gCharHdl;
    hdl[koUml] = &gCharHdl;
    hdl[kuUml] = &gCharHdl;
    hdl[kAUml] = &gCharHdl;
    hdl[kOUml] = &gCharHdl;
    hdl[kUUml] = &gCharHdl;
    hdl[kSZLig] = &gCharHdl;
    hdl[keAcute] = &gCharHdl;

    hdl[kEndChar] = &gEndChar;    
    parFlag = false;
    endHdl = NULL;
}

void
Scanner::translate(Stream *s) {
    unsigned char c;
    List<int> *eh = endHdl;
    endHdl = NULL;
    do {
        c = s->get();
        if (parFlag) hdl[c]->par();
    } while (hdl[c]->translate(s, c));
    // call the end handlers and remove them
    if (endHdl) {
        void (*f)();
        for (f = (void (*)()) endHdl->head(); f; f = (void (*)()) endHdl->next())
            (*f)();
        delete endHdl;
    }
    endHdl = eh;
}

void
Scanner::scan(Stream *s) {
    unsigned char c;
    do {
        c = s->get();
    } while (hdl[c]->scan(s, c));
}

void
Scanner::eatWhite(Stream *s) {
    unsigned char c;
    for(;;) { 
        c = s->get();
        switch (c) {
        case ' ':
        case '\t':
        case '\r':
        case '\n': break;
        case '%': hdl['%']->scan(s, c); break;
	default:
	    s->unget(c);
	    return;
	}
    }
}

void
Scanner::atEnd(void (*f)()) {
    if (! endHdl) endHdl = new List<int>;
    endHdl->insert((int *) f);
}

// character scanner for texts

TextScanner::TextScanner() : Scanner() {
    hdl['\n'] = &whiteHdl;
    hdl['\r'] = &whiteHdl;
    hdl['\t'] = &whiteHdl;
    hdl[' '] = &whiteHdl;
    hdl['#'] = &argHdl;
    hdl['$'] = &mathHdl;
    hdl['&'] = &tabOnlyHdl;
    hdl['~'] = &fixedSpaceHdl;
    //    hdl['_'] = &mathOnlyHdl;
    hdl['^'] = &mathOnlyHdl;
    hdl['%'] = &commentHdl;
    hdl['{'] = &openHdl;
    hdl['}'] = &closeHdl;
    hdl['\\'] = &gCmdChar;
    hdl['-'] = &hyphenHdl;
    hdl['>'] = &gtHTMLHdl;
    hdl['<'] = &ltHTMLHdl;
    hdl['"'] = &quotHTMLHdl;
    hdl[kaUml] = &gUmlChar;
    hdl[koUml] = &gUmlChar;
    hdl[kuUml] = &gUmlChar;
    hdl[kAUml] = &gUmlChar;
    hdl[kOUml] = &gUmlChar;
    hdl[kUUml] = &gUmlChar;
    hdl[kSZLig] = &gUmlChar;
    hdl[keAcute] = &gUmlChar;
}

MathScanner::MathScanner() : Scanner() {
    hdl['\n'] = &whiteHdl;
    hdl['\r'] = &whiteHdl;
    hdl['\t'] = &whiteHdl;
    hdl[' '] = &whiteHdl;
    hdl['#'] = &argHdl;
    hdl['&'] = &tabOnlyHdl;
    hdl['~'] = &fixedSpaceHdl;
    hdl['%'] = &commentHdl;
    hdl['{'] = &openHdl;
    hdl['}'] = &closeHdl;
    hdl['\\'] = &gCmdChar;
    //    hdl['-'] = &hyphenHdl;
    hdl['>'] = &gtHTMLHdl;
    hdl['<'] = &ltHTMLHdl;
    hdl['"'] = &quotHTMLHdl;
    hdl[kaUml] = &gUmlChar;
    hdl[koUml] = &gUmlChar;
    hdl[kuUml] = &gUmlChar;
    hdl[kAUml] = &gUmlChar;
    hdl[kOUml] = &gUmlChar;
    hdl[kUUml] = &gUmlChar;
    hdl[kSZLig] = &gUmlChar;
    hdl[keAcute] = &gUmlChar;
    hdl['$'] = &gEndChar;
    hdl['_'] = &gSubscriptChar;
    hdl['^'] = &gSuperscriptChar;
}

// the plain scanner scans handles braces {...} but passes almost anything 
// else thru

PlainScanner::PlainScanner() : Scanner() {
    hdl['{'] = &openHdl;
    hdl['}'] = &closeHdl;
}

// The Mup scanner is used for commands which allow 
// verbatim $, _ and ^ in their arguments.
MupScanner::MupScanner() : Scanner() {
    hdl['\n'] = &whiteHdl;
    hdl['\r'] = &whiteHdl;
    hdl['\t'] = &whiteHdl;
    hdl[' '] = &whiteHdl;
    hdl['#'] = &argHdl;
    hdl['&'] = &tabOnlyHdl;
    hdl['~'] = &fixedSpaceHdl;
    hdl['%'] = &commentHdl;
    hdl['{'] = &openHdl;
    hdl['}'] = &closeHdl;
    hdl['\\'] = &gCmdChar;
    //    hdl['-'] = &hyphenHdl;
    hdl['>'] = &gtHTMLHdl;
    hdl['<'] = &ltHTMLHdl;
    hdl['"'] = &quotHTMLHdl;
    hdl[kaUml] = &gUmlChar;
    hdl[koUml] = &gUmlChar;
    hdl[kuUml] = &gUmlChar;
    hdl[kAUml] = &gUmlChar;
    hdl[kOUml] = &gUmlChar;
    hdl[kUUml] = &gUmlChar;
    hdl[kSZLig] = &gUmlChar;
    hdl[keAcute] = &gUmlChar;
    hdl['$'] = &gCharHdl;
    hdl['_'] = &gCharHdl;
    hdl['^'] = &gCharHdl;
}


// end of file
