/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// CmdFile.cp -- read command definitions from file

/* Syntax of the command definition file:

def-file ::= [ block ];
block ::= definition [ block ];
definition ::= comment | cmd-def | env-def | def-def | undef-def | ifdef-def | ifndef-def;
comment ::= "%" <any text til end of line>;
cmd-def ::= "cmd" name [ opt-args [ args ] ] [ main-def [ opt-def ] ] ";";
env-def ::= "env" name [ opt-args [ args ] ] [ main-def [ opt-def ] ] [ end-def ] ";";
def-def ::= "def" name [ def-val ] ";";
undef-def ::= "undef" name ";";
ifdef-def ::= "ifdef" name block [ "else" block ] "fi";
ifndef-def ::= "ifndef" name block [ "else" block ] "fi";
name ::= "'" <string of alphabetic characters> "'";
opt-args ::= "1";
args ::= <digit 0-9>;
main-def ::= "'" <string> "'";
opt-def ::= "'" <string> "'";
end-def ::= "'" <string> "'";
def-val ::= "'" <string> "'";

Semantics:
- A command is given by cmd...;, an environment by env...;, a macro definition by def...;,
  a macro is undefined by undef...;, definitions may be given depending on macro values
  by using ifdef...else...fi or ifndef...else...fi
- opt-args gives the number of optional arguments. Currently at most 1 optional 
  argument is supported.
- args gives the number of arguments.
- If neither opt-args nor args are given then both are assumed to be 0.
- If only args is given then opt-args is assumed to be 0.
- If no definition is given then a command or environment is ignored. A macro is
  assigned the empty string as value.
- A command may have a main-def and a opt-def if there is 1 optional argument 
  (opt-arg = 1).
- An actual instance of a TeX command which has one optional arg is replaced by 
  opt-def, if opt-def is given. Otherwise it is replaced by main-def.
- An environment may have a main-def, an end-def and a opt-def if there is 1 
  optional argument (opt-arg = 1).
- The begin of an actual instance of a TeX environment which has one optional arg 
  is replaced by opt-def, if opt-def is given. Otherwise it is replaced by main-def.
- The end of an TeX-environment is replaced by end-def.
- An 'argument pattern' in an definition (which is of the form "#<n>" or 
  "#=<n>" where <n> is either a digit "1"-"9" or the character "a"), is
  replaced by an actual argument of the TeX command or environment.
- An argument pattern "#a" is replaced by an optional argument, if it exists.
  The argument is translated.
- An argument pattern "#=a" is replaced by an optional argument, if it exists.
  The argument is not translated.
- An argument pattern "#<n>" with a digit <n> is replaced by the corresponding
  mandatory argument. The argument is translated.
- An argument pattern "#=<n>" with a digit <n> is replaced by the corresponding
  mandatory argument. The argument is not translated.
- An macro pattern "#(" <string> ")" in a definition is replaced by the macro 
  value string defined by the corresponding def-cmd.
- Macro names may only contain alphabetic characters.
- Macro value strings may not contain # patterns and are taken literally.
- A "'" in a string must be entered as "\'".
- A "#" in a string must be entered as "\#".
- A "\" in a string must be entered as "\\".
- A newline, tab, form-feed, carriage return or backspace may be entered
  like in C as "\n", "\t", "\f", "\r" or "\b".

Examples:
cmd Mexpr 1 '<code>#1</code>';
cmd Mup 1 1 '<code>#1</code>' '<code><a href="\##=a">#1</a></code>';
env 'Mparameter' '<h3>#(params):</h3>';
ifdef 'german'
   def 'params' 'Parameter';
else
   def 'params' 'Parameters';
fi
*/

#include <ctype.h>
#include "stream.h"
#include "mh2html.h"
#include "cmdfile.h"
#include "cmdhdl.h"
#include "mhlpcmds.h"

const int kLineLen = 256;	// max line length for definition file

const int T_end = 0;
const int T_cmd = 1;
const int T_env = 2;
const int T_str = 3;
const int T_num = 4;
const int T_sem = 5;
const int T_err = 6;
const int T_def = 7;
const int T_undef = 8;
const int T_ifdef = 9;
const int T_ifndef = 10;
const int T_else = 11;
const int T_fi = 12;

static void
deferr(FileStream *s) {
  char err[64];
  for(int i=0; i<63; ++i)
    err[i] = s->get();
  err[63] = '\0';
    fatal("illegal command definition file (near '", err, "')");
}

static char
eatWhite(FileStream *s) {
    char c;
    for(;;) {
        c = s->get();
        if (c == ' ' || c == '\t' || c == '\n' || c == '\r') continue;
        if (c != '%') break;
        while ((c = s->get()) != '\n') {	// scan over comments
            if (c == kEndChar) return kEndChar;
        }
    }
    return c;   
}

static int
scanCmdFile(FileStream *s, char *buff, int n) {
    char c = eatWhite(s);
    switch (c) {
    case '\'':	// def string
        while ((c = s->get()) != kEndChar) {
            if (c == '\\') {		// translate masked char
                if (n == 1) return T_err;
                c = s->get();
                switch (c) {
                case kEndChar: return T_err;
                case 'n' : *buff = '\n'; break;
                case 't' : *buff = '\t'; break;
                case 'r' : *buff = '\r'; break;
                case 'b' : *buff = '\b'; break;
                case 'f' : *buff = '\f'; break;
                case '\'' : *buff = '\''; break;
                case '\\' : *buff = '\\'; break;
                case '#' : 
                    if (n == 1) return T_err;
                    *buff = '#'; buff++; n--;
                    *buff = '#'; 
                    break;
                default :
                    if (n == 1) return T_err;
                    *buff = '\\'; buff++; n--;
                    *buff = c;
                }
                buff++; n--;
            } else if (c == '\'') {		// end of string
                *buff = '\0';
                return T_str;
            } else {
                if (n == 1) return T_err;
                *buff = c; buff++; n--;
            }
        }
        return T_err;
    case kEndChar:
        return T_end;
    case ';':
        return T_sem;
    case 'c':
        c = s->get();
        if (c != 'm') return T_err;
        c = s->get();
        if (c != 'd') return T_err;
        return T_cmd;
    case 'd':
        c = s->get();
        if (c != 'e') return T_err;
        c = s->get();
        if (c != 'f') return T_err;
        return T_def;
    case 'e':
        c = s->get();
        if (c == 'l') {
	    c = s->get();
	    if (c != 's') return T_err;
	    c = s->get();
	    if (c != 'e') return T_err;
            return T_else;
        }
        if (c != 'n') return T_err;
        c = s->get();
        if (c != 'v') return T_err;
        return T_env;
    case 'f':
        c = s->get();
        if (c != 'i') return T_err;
        return T_fi;
    case 'i':
        c = s->get();
        if (c != 'f') return T_err;
        c = s->get();
        if (c == 'n') {
	    c = s->get();
	    if (c != 'd') return T_err;
	    c = s->get();
	    if (c != 'e') return T_err;
	    c = s->get();
	    if (c != 'f') return T_err;
            return T_ifndef;
        } 
        if (c != 'd') return T_err;
        c = s->get();
        if (c != 'e') return T_err;
        c = s->get();
        if (c != 'f') return T_err;
        return T_ifdef;
    case 'u':
        c = s->get();
        if (c != 'n') return T_err;
        c = s->get();
        if (c != 'd') return T_err;
        c = s->get();
        if (c != 'e') return T_err;
        c = s->get();
        if (c != 'f') return T_err;
        return T_undef;
    }
    if (isdigit(c)) {
        buff[0] = c;
        buff[1] = '\0';
        return T_num;
    }
    return T_err;
}

static void
parseCmd(FileStream *s, bool isEnv, bool ignore) {
    int t, opts, args;
    static char nam[128], def1[1024], def2[1024], def3[1024], dummy[2];
    
    // get command/env name
    t = scanCmdFile(s, nam, 128);
    if (t != T_str) deferr(s);
    // parse arg numbers
    t = scanCmdFile(s, def1, 1024);
    if (t == T_num) {
        sscanf(def1, "%d", &opts);
        if (opts < 0 || opts > 9) deferr(s);
        t = scanCmdFile(s, def1, 1024);
        if (t == T_num) {
            sscanf(def1, "%d", &args);
            if (args < 0 || args > 9) deferr(s);
            if (opts != 1) deferr(s);
            t = scanCmdFile(s, def1, 1024);
        } else {
            args = opts; opts = 0;
        }
    } else {
        opts = 0; args = 0;
    }
    // non-empty definition?
    if (t == T_sem) {
        if (ignore) return;
        if (isEnv) {
            ParsedEnv *e = new ParsedEnv(opts, args, NULL, NULL, NULL);
            if (! e) deferr(s);
            gEnvs.push(nam, e);
        } else {
            ParsedCmd *e = new ParsedCmd(opts, args, NULL, NULL);
            if (! e) deferr(s);
            gCmds.push(nam, e);
        }
        return;
    }
    if (t == T_str) t = scanCmdFile(s, def2, 1024);
    else deferr(s);
    // only main definition?
    if (t == T_sem) {
        if (ignore) return;
        if (isEnv) {
            ParsedEnv *e = new ParsedEnv(opts, args, def1, NULL, NULL);
            if (! e) deferr(s);
            gEnvs.push(nam, e);
        } else {
            ParsedCmd *e = new ParsedCmd(opts, args, def1, NULL);
            if (! e) deferr(s);
            gCmds.push(nam, e);
        }
        return;
    } else if (! isEnv && opts == 0) deferr(s);
    if (t == T_str) t = scanCmdFile(s, def3, 1024);
    else deferr(s);
    // optional definition or end definition?
    if (t == T_sem) {
        if (ignore) return;
        if (isEnv) {
            ParsedEnv *e = new ParsedEnv(opts, args, def1, NULL, def2);
            if (! e) deferr(s);
            gEnvs.push(nam, e);
        } else {
            ParsedCmd *e = new ParsedCmd(opts, args, def1, def2);
            if (! e) deferr(s);
            gCmds.push(nam, e);
        }
        return;
    } else if ((! isEnv) || (isEnv && opts == 0)) deferr(s);
    
    if (t == T_str) t = scanCmdFile(s, dummy, 2);
    else deferr(s);
    if (t != T_sem) deferr(s);
    if (ignore) return;
    ParsedEnv *e = new ParsedEnv(opts, args, def1, def2, def3);
    if (! e) deferr(s);
    gEnvs.push(nam, e);
}

static void
parseDef(FileStream *s, bool ignore) {
    int t;
    static char *v, nam[128], def[1024];
    
    // get macro name
    t = scanCmdFile(s, nam, 128);
    if (t != T_str) deferr(s);
    // parse macro value
    t = scanCmdFile(s, def, 1024);
    if (t == T_str) {
        if (! ignore) {
            v = new char[strlen(def)+1];
            if (! v) deferr(s);
            strcpy(v, def);
        }
    } else if (t == T_sem) {
        if (! ignore) {
            v = new char[1];
            if (! v) deferr(s);
            *v = '\0';
        }
    } else {
        deferr(s);
    }
    if (ignore) return;
    // remove old values and push new one
    while (gDefs.top(nam)) gDefs.pop(nam);
    // special definition?
    if (strcmp(nam, "german") == 0) {
        gLanguage = kGerman;
    } else if (strcmp(nam, "english") == 0) {
        gLanguage = kEnglish;
    } else if (strcmp(nam, "ie4") == 0) {
        gIe4 = true;
    }
    gDefs.push(nam, v);
}

static void
parseUndef(FileStream *s, bool ignore) {
    int t;
    static char nam[128], dummy[2];
    
    // get macro name
    t = scanCmdFile(s, nam, 128);
    if (t != T_str) deferr(s);
    // parse ;
    t = scanCmdFile(s, dummy, 2);
    if (t != T_sem) deferr(s);
    if (ignore) return;
    // remove values
    while (gDefs.top(nam)) gDefs.pop(nam);
}

static int parseBlock(FileStream *, bool);

static void
parseIf(FileStream *s, bool isNdef, bool ignore) {
    int t; 
    bool noIf, noElse;
    static char nam[128];
    
    // get macro name
    t = scanCmdFile(s, nam, 128);
    if (t != T_str) deferr(s);
    // check if defined
    if (ignore) {
        noIf = noElse = true;
    } else {
        noIf = (gDefs.top(nam) == NULL);
        if (isNdef) noIf = (! noIf);
        noElse = (! noIf);
    }
    // parse if block
    t = parseBlock(s, noIf);
    if (t == T_else) {
        // parse else block
        t = parseBlock(s, noElse);
    }
    if (t != T_fi) deferr(s);
}

static int
parseBlock(FileStream *s, bool ignore) {
    int t;
    static char buff[128];

    for (;;) {
        t = scanCmdFile(s, buff, 128);
        switch (t) {
        case T_cmd:
            parseCmd(s, false, ignore);
            break;
        case T_env:
            parseCmd(s, true, ignore);
            break;
        case T_def:
            parseDef(s, ignore);
            break;
        case T_undef:
            parseUndef(s, ignore);
            break;
        case T_ifdef:
            parseIf(s, false, ignore);
            break;
        case T_ifndef:
            parseIf(s, true, ignore);
            break;
        case T_sem:
            break;
        default:
            return t;
        }
    }
}

void initCmdsFromFile(FileStream *s, char *sname) {

    if (parseBlock(s, false) != T_end) deferr(s);
}

// end of file
