/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */


// MhlpEnvs.cp -- special Mhelp.sty environments

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#include "files.h"
#include "mh2html.h"
#include "charhdl.h"
#include "basecmds.h"
#include "mhlpcmds.h"
#include "mhlpenvs.h"
#include "counter.h"
#include "utils.h"

#ifdef sun
#include <alloca.h>
#endif

#ifdef WIN32
#include <assert.h>
#include <direct.h>
#include <malloc.h>
#define alloca _alloca
#endif

const int kArgLen = 2048;	// max length of command arguments

// ItemizeEnv is a base class for itemizing environments

class ItemizeEnv : public EnvHdl {
public:
    ItemizeEnv(ParaType pt, int o, int a) : EnvHdl(o, a), para_type(pt) {}
    virtual bool translateBody(Stream *s);
protected:
    ParaType para_type;
};

bool
ItemizeEnv::translateBody(Stream *s) {
    ParaType pt = gParaType;
    gParaType = para_type;
    bool res = EnvHdl::translateBody(s);
    gParaType = pt;
    return res;
}

// ParsedItemizeEnv is a abstract subclass of ParsedEnv which
// does itemizing

class ParsedItemizeEnv : public ParsedEnv {
public:
    ParsedItemizeEnv(ParaType pt, int no, int na, char *md, char *od, char *ed)
        : ParsedEnv(no, na, md, od, ed), para_type(pt) {}
    virtual bool translateBody(Stream *s);
protected:
    ParaType para_type;
};

bool
ParsedItemizeEnv::translateBody(Stream *s) {
    ParaType pt = gParaType;
    gParaType = para_type;
    bool res = ParsedEnv::translateBody(s);
    gParaType = pt;
    return res;
}

// SectionHeaderEnv is the base class for section headers

class SectionHeaderEnv : public ItemizeEnv {
public:
    SectionHeaderEnv(const char* t, ParaType pt, int o, int a) 
        : ItemizeEnv(pt, o, a), title(t) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateTitle(Stream *, SubStream *[], int, SubStream *[]);
    static void reset();
protected:
    const char* title;
    static int num;
};

int SectionHeaderEnv::num = 1;

void SectionHeaderEnv::reset()
{
    num = 1;
}

bool 
SectionHeaderEnv::translateTitle(Stream *s, SubStream *oa[], int no, SubStream *a[])
{
    assert(this);
    Puts(gDefs.top((char*) title));
    return true;
}

bool 
SectionHeaderEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[])
{
    if (gIe4) {
        char sect[64];
        sprintf(sect, "sct%d", num);
        char aSect[64];
        sprintf(aSect, "aSct%d", num);
        num++;

        Puts("<h4 class=\"sect\"><a onclick=\"expd(");
        Puts(sect);
        Puts(",");
        Puts(aSect);
        Puts(")\" href=\"#Z\">\n<img src=\"../PICTURES/open.gif\" id=\"");
        Puts(aSect);
        Puts("\" border=0>");
        if (! translateTitle(s, oa, no, a))
            return false;
        Puts("</a></h4>\n<div class=\"exp\" id=\"");
        Puts(sect);
        Puts("\" style=\"display:;\">");
    } else {
        Puts("<h4 class=\"sect\"><img src=\"../PICTURES/closed.gif\" border=0> ");
        if (! translateTitle(s, oa, no, a))
            return false;
        Puts("</h4>");
    }
    switch (para_type) {
    case kItem:
        Puts("<ul>");
        break;
    case kTableEntry:
    case kLastTableEntry:
        Puts("<table border=0>");
	gScanner->donePar();
        break;
    default:
      break;
    }
    return true;
}

bool
SectionHeaderEnv::translateEnd(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    switch (para_type) {
    case kItem:
        Puts("</ul>");
        break;
    case kTableEntry:
    case kLastTableEntry:
        Puts("</table>");
        break;
    default:
      break;
    }
    if (gIe4)
        Puts("</div>");
    return ItemizeEnv::translateEnd(s, oa, no, a);
}

// special verbatim env for direct input of HTML code
class MhtmlEnv : public VerbatimEnv {
public:
  MhtmlEnv(int oargs, int args, char *end) :  VerbatimEnv(oargs, args, end) {}
  //  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
  virtual bool translateBody(Stream*);
} mhtmlEnv(0, 0, "\\end{Mhtml}");

bool 
MhtmlEnv::translateBody(Stream *s) {
    char c;
    const char *em = endMark, *cp;
    Scanner *pS=gScanner;
    gScanner=&gPlainScanner;
    //    Puts("<pre>");
    while (isspace(c = s->get()) && '\n'!=c) /* nix */;
    if('\n'==c) c=s->get();
    for(;;) {
        if (c == kEndChar)
            s->fatal("run-away Mhtml environment");
	if (c == *em) {
	    em++;
	    if (*em == '\0') break;
	} else {
	    for (cp = endMark; cp != em; cp++) Putc(*cp);
	    em = endMark;
	    if (c == *em) em++; 
	    else Putc(c);
	}
        c = s->get();
    }
    //    Puts("</pre>");
    gCurrEnvName.pop();
    gScanner=pS;
    return true;
}

// The MChanges environment

static
class MChangesVersionCls : public CmdHdl {
public:
  MChangesVersionCls(const char * const _version) 
    : CmdHdl (0, 1), print(false), version(_version) {}
  virtual bool translateArgs( Stream *s, SubStream *oa[],
			      int no, SubStream *a[]) {
    char *given_version=a[0]->cont();
    print = (strcmp(version,given_version)==0);
    if(print) {
      Puts("<h4 class=\"sect\">");
      Puts(gDefs.top("changes"));
      Puts("</h4>\n<ul>\n");
    }
    return true;
  }
  bool print;
protected:
  const char * const version;
} MChangesVersion("3.0");

static
class MChangesEngCls : public CmdHdl {
public:
  MChangesEngCls(MChangesVersionCls *vercont) 
    : CmdHdl(1,1), version_controller(vercont) {}
  virtual bool translateArgs( Stream *s, SubStream *oa[],
			      int no, SubStream *a[]) {
    if(version_controller->print && gLanguage == kEnglish) {
        putLangPara(a[0]);
    }
    return true;
  }
protected:
  MChangesVersionCls * version_controller;
} MChangesEng(&MChangesVersion);

static
class MChangesDeCls : public CmdHdl {
public:
  MChangesDeCls(MChangesVersionCls *vercont)
    : CmdHdl(1,1), version_controller(vercont) {}
  virtual bool translateArgs( Stream *s, SubStream *oa[],
			      int no, SubStream *a[]) {
    if(version_controller->print && gLanguage == kGerman) {
        putLangPara(a[0]);
    }
    return true;
  }
protected:
  MChangesVersionCls * version_controller;
} MChangesDe(&MChangesVersion);

static
class MChangesNewCls : public CmdHdl {
public:
  MChangesNewCls(MChangesVersionCls *vercont)
    : CmdHdl(1,0), version_controller(vercont) {}
  virtual bool translateArgs( Stream *s, SubStream *oa[],
			      int no, SubStream *a[]) {
    if(version_controller->print) {
      Puts("<li>");
      if(no) {
        Puts("<code>");
        Puts(oa[0]->cont());
        Puts("</code>");
      } else {
        gCmds.top("Mthis")->translateArgs(s, NULL, 0, NULL);
      }
      Puts(" ");
      Puts(gDefs.top("new"));
      Puts("</li>\n");
    }
    return true;
  }
protected:
  MChangesVersionCls * version_controller;
} MChangesNew(&MChangesVersion);

static
class MChangesRenamedCls : public CmdHdl {
public:
  MChangesRenamedCls(MChangesVersionCls *vercont)
    : CmdHdl(1,1), version_controller(vercont) {}
  virtual bool translateArgs( Stream *s, SubStream *oa[],
			      int no, SubStream *a[]) {
    if(version_controller->print) {
      Puts("<li>");
      if(no) {
        Puts("<code>");
        Puts(oa[0]->cont());
        Puts("</code>");
      } else {
        gCmds.top("Mthis")->translateArgs(s, NULL, 0, NULL);
      }
      Puts(" ");
      Puts(gDefs.top("renamed"));
      Puts(" <code>");
      gScanner->translate(a[0]);
      Puts("</code>.");
      Puts("</li>\n");
    }
    return true;
  }
protected:
  MChangesVersionCls * version_controller;
} MChangesRenamed(&MChangesVersion);

static
class MChangesNoneCls : public CmdHdl {
public:
  MChangesNoneCls(MChangesVersionCls *vercont)
    : CmdHdl(0,0), version_controller(vercont) {}
  virtual bool translateArgs( Stream *s, SubStream *oa[],
			      int no, SubStream *a[]) {
    if(version_controller->print) {
      Puts("<li>");
      Puts(gDefs.top("nochange"));
      Puts("</li>\n");
    }
    return true;
  }
protected:
  MChangesVersionCls * version_controller;
} MChangesNone(&MChangesVersion);

class MChangesEnv : public EnvHdl {
public:
  MChangesEnv(CmdHdl *_eng, CmdHdl *_de,
              CmdHdl *_new, CmdHdl *_none,
              CmdHdl *_renamed,
	      MChangesVersionCls *_version) 
    : EnvHdl(0,0), eng(_eng), de(_de), new_(_new),
      none(_none), renamed(_renamed), version(_version) {}
  virtual bool translateBody(Stream *s);
protected:
  CmdHdl *eng, *de, *new_, *none, *renamed;
  MChangesVersionCls *version;
} mchangesEnv(&MChangesEng, &MChangesDe, &MChangesNew,
              &MChangesNone, &MChangesRenamed, &MChangesVersion);

bool
MChangesEnv::translateBody(Stream *s) {
  gCmds.push("eng",     eng);
  gCmds.push("de",      de);
  gCmds.push("new",     new_);
  gCmds.push("none",    none);
  gCmds.push("renamed", renamed);
  gCmds.push("version", version);

  version->print = false;
  ParaType sp = gParaType;
  gParaType = kItem;
  gScanner->translate(s);
  gParaType = sp;

  Puts("\n</ul>\n");
  gCmds.pop("version");
  gCmds.pop("renamed");
  gCmds.pop("new");
  gCmds.pop("none");
  gCmds.pop("de");
  gCmds.pop("eng");

  return true;
}

// special verbatim envs for MuPAD input and output

class MxcallEnv : public VerbatimEnv {
public:
    MxcallEnv(int o, int a, char *e) : VerbatimEnv(o, a, e) {}
    virtual bool translateBody(Stream *);
}; // mxcallEnv(0, 0, "\\end{Mxcall}");

bool 
MxcallEnv::translateBody(Stream *s) {
    char c;
    bool nl = false;
    const char *em = endMark, *cp;
    Puts("<pre class=\"MuPAD input\"><span class=\"MuPAD Prompt\">&gt;&gt;</span> ");
    while (isspace(c = s->get())) /* nix */;
    for(;;) {
        if (c == kEndChar)
            s->fatal("run-away verbatim environment");
	if (c == *em) {
	    em++;
	    if (*em == '\0') break;
	} else {
	    if (nl) { Puts("   "); nl = false; }
	    for (cp = endMark; cp != em; cp++) PutVerbc(*cp);
	    em = endMark;
	    if (c == *em) em++; 
	    else {
	        PutVerbc(c);
	        if (c == '\n') nl = true;
	    }
	}
        c = s->get();
    }
    Puts("</pre>");
    gCurrEnvName.pop();
    return true;
}

class MxresultEnv : public VerbatimEnv {
public:
    MxresultEnv(int o, int a, char *e) : VerbatimEnv(o, a, e) {}
    virtual bool translateBody(Stream *);
}; // mxresultEnv(0, 0, "\\end{Mxresult}");

bool 
MxresultEnv::translateBody(Stream *s) {
    char c;
    bool nl = false;
    const char *em = endMark, *cp;
    Puts("<pre>      ");
    while (isspace(c = s->get()) && c != '\n') /* nix */;
    if( c == '\n' ) c = s->get();
    for(;;) {
        if (c == kEndChar)
            s->fatal("run-away verbatim environment");
	if (c == *em) {
	    em++;
	    if (*em == '\0') break;
	} else {
	    if (nl) { Puts("      "); nl = false; }
	    for (cp = endMark; cp != em; cp++) PutVerbc(*cp);
	    em = endMark;
	    if (c == *em) em++; 
	    else {
	        PutVerbc(c);
	        if (c == '\n') nl = true;
	    }
	}
        c = s->get();
    }
    Puts("</pre>");
    gCurrEnvName.pop();
    return true;
}

static class MvinEnv : public MxcallEnv {
protected:
    bool hide;
public:
    MvinEnv(int a, char *e) : MxcallEnv(0, a, e) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateBody(Stream *);
} mvinEnv(1, "\\end{Mvin}");//, mvinnEnv(1, "\\end{Mvinn}");

bool
MvinEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
  std::vector<std::string> args = split_string(a[0]->cont(), ',');
  std::vector<std::string>::iterator pos;

  hide=false;
  for(pos = args.begin(); pos != args.end(); ++pos) {
    if((*pos)=="hide" || (*pos)=="hide=true")
      hide = true;
  }
  return true;
}

bool 
MvinEnv::translateBody(Stream *s) {
    return (hide ? scanBody(s) : MxcallEnv::translateBody(s));
}

static class MexinEnv : public MxcallEnv {
protected:
    bool hide;
public:
    MexinEnv(int a, char *e) : MxcallEnv(0, a, e), hide(false) {}
    virtual bool translateBody(Stream *);
} mexinEnv(0, "\\end{Mexin}");


bool 
MexinEnv::translateBody(Stream *s) {
    return (hide ? scanBody(s) : MxcallEnv::translateBody(s));
}

static class MvoutEnv : public MxresultEnv {
protected:
    bool hide;
public:
    MvoutEnv(char *e) : MxresultEnv(0, 1, e) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateBody(Stream *);
} mvoutEnv("\\end{Mvout}");

bool
MvoutEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
  std::vector<std::string> args = split_string(a[0]->cont(), ',');
  std::vector<std::string>::iterator pos;

  hide=false;
  for(pos = args.begin(); pos != args.end(); ++pos) {
    if((*pos)=="hide" || (*pos)=="hide=true")
      hide = true;
  }
  return true;
}

bool
MvoutEnv::translateBody(Stream *s) {
    return (hide ? scanBody(s) : MxresultEnv::translateBody(s));
}

static class MexoutEnv : public MxresultEnv {
protected:
    bool hide;
public:
    MexoutEnv(char *e) : MxresultEnv(0, 0, e), hide(false) {}
    virtual bool translateBody(Stream *);
} mexoutEnv("\\end{Mexout}");


bool 
MexoutEnv::translateBody(Stream *s) {
    return (hide ? scanBody(s) : MxresultEnv::translateBody(s));
}

// Mpackage defines the actual package name gMpack

static class MpackageEnv : public ParsedEnv {
public:
    MpackageEnv() :
        ParsedEnv(0, 2, 
          "<h1><code><a name=\"#L1\" class=\"Malias\">#1</a></code> -- #2</h1>", 
          NULL, NULL) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
} mpackageEnv;

bool 
MpackageEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
  char *buf;

  if( -1 ==
#ifdef WIN32
	  _mkdir( a[0]->cont() )
#else
	  mkdir( a[0]->cont(), 0777 )
#endif
		) {
    switch (errno) {
    case EEXIST: break;
    default: 
      a[0]->fatal("Can't make directory \"", a[0]->cont(), "\"");
      break;
    }
  }
  buf = reinterpret_cast<char*>(alloca ( strlen(a[0]->cont()) 
					+ sizeof(".//index") ));
  sprintf( buf, "./%s/index", a[0]->cont() );
  OpenHTMLFile( buf );
  PutHeader(s);

    SectionHeaderEnv::reset();

    // print title into contents file
    if (OpenNewContFile()) {
      //        PutConts("\\tocpageline{2}{");
      //        PutConts(GetHTMLFileName());
      PutConts("\\Mhtmlverb{<h1>}\\Mhref[index]{\\Mex{");
        PutConts(a[0]->cont());
        PutConts("} -- ");
        PutConts(a[1]->cont());
        PutConts("}\\Mhtmlverb{</h1>}\n");
	PutConts("\\Mhtmlverb{<h2>}{\\eng{Table of contents}");
	PutConts("\\de{Inhaltsverzeichnis}}\\Mhtmlverb{</h2>}\n");
    } else s->error("could not open contents file");

    gMpack = a[0];
    return ParsedEnv::translateArgs(s, oa, no, a);
}

bool
MpackageEnv::translateEnd(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    gMpack = &gNilStream;
    ParsedEnv::translateEnd(s, oa, no, a);
    PutFooter(s);
    CloseHTMLFile();
    return true;
}

// Msyntax seperates Margs items

static class MsyntaxEnv : public ParsedEnv {
public:
    MsyntaxEnv() :
        ParsedEnv(0, 0,
	  "<h4 class=\"sect\">#(callseq):</h4>\n",
          NULL, "") {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
} msyntaxEnv;

bool 
MsyntaxEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    gInSyntax = true;
    return ParsedEnv::translateArgs(s, oa, no, a);
}

bool
MsyntaxEnv::translateEnd(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    gInSyntax = false;
    return ParsedEnv::translateEnd(s, oa, no, a);
}

// MCalls, MDomConst, MDomElConst

static class MCallsEnvItemCmd : public CmdHdl {
public:
  MCallsEnvItemCmd() : CmdHdl(0, 0) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mcallsEnvItemCmd;

bool
MCallsEnvItemCmd::translateArgs(Stream *s, SubStream *oa[],
				int no, SubStream *ar[]) {
  Puts("<br>\n");
  return true;
}

static class MCallsEnv : public SectionHeaderEnv {
public:
    MCallsEnv(const char * const heading) : 
      SectionHeaderEnv(heading, kPara, 0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateBody(Stream *s);
} mCallsEnv("callseq"), mDomConstEnv("domconst"), 
  mDomElConstEnv("domelconst"), mAxConstEnv("axconst"),
  mCatConstEnv("catconst");

bool 
MCallsEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    if (! SectionHeaderEnv::translateArgs(s, oa, no, a))
        return false;
    Puts("<p>");
    gScanner->donePar();
    gCmds.push("item", &mcallsEnvItemCmd);
    return true;
}

bool
MCallsEnv::translateEnd(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    Puts("</p>");
    gCmds.pop("item");
    return SectionHeaderEnv::translateEnd(s, oa, no, a);
}

bool
MCallsEnv::translateBody(Stream *s) {
  localScanner _(&gMupScanner);
  return SectionHeaderEnv::translateBody(s);
}

// Various standard section headers

static class MSectEnv : public SectionHeaderEnv {
public:
    MSectEnv(const char *headid) : SectionHeaderEnv(headid, kPara, 0, 0) {}
} mReturnValueEnv("retval"), mSideEffectsEnv("sideeffects"), 
  mFurtherDocEnv("furtherdoc"), mOperationsEnv("operations"),
  mEvalEnv("secteval"), mFuncCallEnv("funccall"), mEnvSensEnv("envsens"),
  mOperandsEnv("operands"), mOutputEnv("output"), 
  mOverloadingEnv("overload");

static class MTableSectEnv : public SectionHeaderEnv {
public:
  MTableSectEnv(const char *headid) : SectionHeaderEnv(headid, kTableEntry, 0, 0) {}
  virtual bool translateBody(Stream *);
} mParametersEnv("parameters"), mOptionsEnv("options");

bool MTableSectEnv::translateBody(Stream *s) {
  bool res;
  Puts("<table>\n");
  res = SectionHeaderEnv::translateBody(s);
  Puts("</table>\n");
  return res;
}

// MOptionDetailsEnv describes the details of an option

static class MOptionDetailsEnv : public SectionHeaderEnv {
public:
    MOptionDetailsEnv() 
        : SectionHeaderEnv("option", kItem, 0, 1) {}
    virtual bool translateTitle(Stream *, SubStream *[], int, SubStream *[]);
} mOptionDetailsEnv;

bool 
MOptionDetailsEnv::translateTitle(Stream *s, SubStream *oa[], int no, SubStream *a[])
{
    if (! SectionHeaderEnv::translateTitle(s, oa, no, a))
        return false;
    Puts(": ");
    gScanner->translate(a[0]);
    return true;
}

// MExample describes an example

Counter gExampleCounter;
static string currHelppage;

static class MExampleEnv : public SectionHeaderEnv {
public:
    MExampleEnv() : SectionHeaderEnv("example", kItPara, 0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateTitle(Stream *, SubStream *[], int, SubStream *[]);
} mExampleEnv;

bool
MExampleEnv::translateTitle(Stream *s, SubStream *oa[], int no, SubStream *a[])
{
  char buff[10];
  Puts(gDefs.top((char*) title));
  sprintf(buff, " %d", gExampleCounter.get());
  Puts(buff);
  return true;
}

bool 
MExampleEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    ++gExampleCounter;
    Puts("<a name=\"");
//     if (! gMpack.empty() 
// 	&& strcmp(gMpack.cont(),"stdlib") != 0
// 	&& strcmp(gMpack.cont(),"datatypes") != 0
// 	) { 
//       gScanner->translate(&gMpack);
//       gMpack.reset();
//       Puts("::"); 
//     }
//     gScanner->translate(&gMname);
//     gMname.reset();
//     Puts("::");
    Puts(a[0]->cont());
    Puts("\"></a>");
    a[0]->reset();
    if (! SectionHeaderEnv::translateArgs(s, oa, no, a))
        return false;

    gLocalLables[std::make_pair(currHelppage,std::string(a[0]->cont()))] 
      = gExampleCounter.get();
    PutLocalLabel(std::string(a[0]->cont()),gExampleCounter.get());

    return true;
}

// MDetails describes the details of a function

static class MDetailsEnv : public SectionHeaderEnv {
public:
    MDetailsEnv()
        : SectionHeaderEnv("description", kItem, 0, 0) {}
} mDetailsEnv;

// MBackground lists background information

static class MBackgroundEnv : public SectionHeaderEnv {
public:
    MBackgroundEnv()
        : SectionHeaderEnv("background", kItem, 0, 0) {}
} mBackgroundEnv;

// Mhelppage defines a new style help page

Counter gHelppageCounter;

static class MhelppageEnv : public EnvHdl {
public:
    MhelppageEnv() : EnvHdl(1, 2) {}
  virtual bool translate(Stream*);
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
} mhelppageEnv;

bool
MhelppageEnv::translate(Stream *s) {
    int i, no;
    bool cont;
    SubStream *a;
    SubStream *opts[9];
    SubStream *args[9];

    for (i = 0; i < 9; i++) opts[i] = NULL;

    {
      localScanner _(&gMupScanner);
      // get optional args
      for (no = 0, i = 0; i < nopt; i++) {
        a = getOptArg(s);
        if (a) { opts[i] = a; no++; }
        else break;
      }
      args[0] = getArg(s);
    }
    // get other args
    for (i = 1; i < narg; i++) {
        a = getArg(s);
        if (a) args[i] = a;
        else { 
            s->error("argument(s) missing"); 
            return true;
        }
    }
    for (i = narg; i < 9; i++) args[i] = NULL;
    // handle command
    cont = translateArgs(s, opts, no, args);
    if (cont) cont = translateBody(s);
    if (cont) cont = translateEnd(s, opts, no, args);
    // delete args lists
    for (i = 0; i < no; i++) delete opts[i];
    for (i = 0; i < narg; i++) delete args[i];
    return cont;
}

bool 
MhelppageEnv::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    gMname = a[0];
    gAuthor = a[1];

    currHelppage = a[0]->cont();
    // eject page
    //    PutFooter(s); // is done in OpenNextHTMLFile
    OpenNextHTMLFile(s);
    //    PutHeader(s); // dito
    SectionHeaderEnv::reset();

    ++gHelppageCounter;

    if (no == 0) {
        gMnameLabel = &gNilStream;
        PutLabelAlias(gMname.cont());
    } else {
        gMnameLabel = oa[0];
        PutLabelAlias(gMnameLabel.cont());
    }
    return true;
}

bool
MhelppageEnv::translateEnd(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    gMname = &gNilStream;
    gMnameLabel = &gNilStream;
    gAuthor = &gNilStream;
    return EnvHdl::translateEnd(s, oa, no, a);
}

// MInfo defines the title for a help page

class MCommaListHandler : public CharHdl {
private:
  const char *const prefix;
public:
  MCommaListHandler(const char *const pref) :
    CharHdl(), prefix(pref) {}
  virtual bool translate(Stream *s, char) {
    char c;
    c = s->get();
    while(isspace(c) || c == ',')
      c = s->get();
    Puts(", ");
    if (kEndChar != c) {
      s->unget(c);
      if(prefix) {
	Puts(prefix);
	Puts("::");
      }
    }
    return true;
  }
};

class MCommaListScanner : public MupScanner {
private:
  MCommaListHandler hdl_;
public:
  MCommaListScanner(const char *const pref) 
    : MupScanner(), hdl_(pref) {
    hdl[','] = &hdl_;
    if(pref) {
      Puts(pref);
      Puts("::");
    }
  }
};

static class MInfoEnv : public EnvHdl {
public:
    MInfoEnv() : EnvHdl(0, 0) {}
    virtual bool translateBody(Stream *);
} mInfoEnv;

bool MInfoEnv::translateBody(Stream *s) {
    bool cont;
    char *prefix = NULL;

    // put title
    Puts("<h2 class=\"Mhelppage\"><code><a name=\"");
    if (! gMpack.empty() 
	&& strcmp(gMpack.cont(),"stdlib") != 0
	&& strcmp(gMpack.cont(),"datatypes") != 0
	) { 
      gScanner->translate(&gMpack);
      gMpack.reset();
      Puts("::"); 
    }
    if (*(gMnameLabel.cont()) != 0) {
        Puts(gMnameLabel.cont());
    } else {
        Puts(gMname.cont());
    }
    Puts("\">");
    if (! gMpack.empty() 
	&&  strcmp ( gMpack.cont(), "stdlib" ) != 0
	&&  strcmp ( gMpack.cont(), "datatypes" ) != 0
	) {
      /*      gScanner->translate(&gMpack);
      gMpack.reset();
      Puts("::"); */
      prefix = gMpack.cont();
    }
    {
      MCommaListScanner localscan(prefix);
      localScanner _(&localscan);
      localscan.translate(&gMname);
    }
    gMname.reset();
    Puts("</a></code> -- ");
    ParaType pt = gParaType;
    gParaType = kNone;
    StreamPos p = s->pos();
    cont = EnvHdl::translateBody(s);
    gParaType = pt;
    Puts("</h2>\n");

    // put author
    if (gPutVersion) {
        Puts("<cite>");
        gScanner->translate(&gAuthor);
        gAuthor.reset();
        Puts("</cite>\n");
    }

    // get title for use in index files
    char buff[kArgLen];
    if (! s->getRange(p, buff, kArgLen))
	s->fatal("body too long");

    // printe title into contents file
    if (OpenNewContFile()) {
        PutConts("\\tocpageline{1}{");
        PutConts(GetHTMLFileName());
        PutConts("}{\\Mex{");
        PutConts(gMname.cont());
        PutConts("}}{\\begin{MInfo}");
        PutConts(buff);
        PutConts("}\n");
    } else s->error("could not open contents file");

    // print toc line
    if (OpenNewTocFile()) {
        PutTocs("\\Mtocline");
	if (*(gMnameLabel.cont()) != 0) {
	    PutTocs("[");
	    PutTocs(gMnameLabel.cont());
	    PutTocs("]");
	}
        PutTocs("{");
        PutTocs(gMname.cont());
        PutTocs("}{\\begin{MInfo}");
        PutTocs(buff);
        PutTocs("}\n");
    } else s->error("could not open toc file");

    return cont;
}

// Mseealso defines objects related to the current one

static class MseealsoEnv : public ParsedEnv {
public:
    MseealsoEnv() 
        : ParsedEnv(0, 0, "<h4 class=\"sect\">#(seealso):</h4>", NULL, NULL) {}
} mseealsoEnv;

// Moverloading defines arguments which may cause overloading

static class MoverloadingEnv : public ParsedEnv {
public:
    MoverloadingEnv() 
        : ParsedEnv(0, 0, "<h4 class=\"sect\">#(overloadable):</h4>", NULL, NULL) {}
} moverloadingEnv;

// Mgerman ignores its body if Mlangenglish is stated,
// otherwise Umlaute may be given via " (instead of \")

static class MgermanEnv : public EnvHdl {
public:
    MgermanEnv() : EnvHdl(0, 0) {}
    virtual bool translateBody(Stream *);
} mgermanEnv;

bool 
MgermanEnv::translateBody(Stream *s) {
    bool res;
    //    CharHdl *h = gScanner->getHdl('"');
    //    gScanner->setHdl('"', &gGermanQuoteChar);
    if (gLanguage == kGerman)
        res = EnvHdl::translateBody(s);
    else
        res = EnvHdl::scanBody(s);
    //    gScanner->setHdl('"', h);
    return res;
}

// Menglish ignores its body if Mlanggerman is stated

static class MenglishEnv : public EnvHdl {
public:
    MenglishEnv() : EnvHdl(0, 0) {}
    virtual bool translateBody(Stream *);
} menglishEnv;

bool 
MenglishEnv::translateBody(Stream *s) {
    if (gLanguage == kEnglish) 
        return EnvHdl::translateBody(s);
    else
        return EnvHdl::scanBody(s);
}

// MMethodDetails initializes internal commands which need the parameter
static class MMethodDetailsEnv : public EnvHdl {
public:
  bool inlist;
  SubStream fn;

  MMethodDetailsEnv() : EnvHdl(0, 1), inlist(false) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
  virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
  //  virtual bool scanArgs(Stream *, SubStream *[], int, SubStream *[]);
  //  virtual bool scanEnd(Stream *, SubStream *[], int, SubStream *[]);
} mMethodDetailsEnv;

// helper classes for call, eng, and de inside MMethodDetails

class McallCmdMMethDet : public CmdHdl {
protected:
  MMethodDetailsEnv * parent;
public:
  McallCmdMMethDet(MMethodDetailsEnv *p) : CmdHdl(0,1), parent(p) {}
  virtual bool translateArgs(Stream *s, SubStream *oa[],
			     int no, SubStream *a[]) {
    if(parent->inlist)
      Puts("</li></ul>\n");
    Puts("<li>");
    Scanner *sc = gScanner;
    gScanner = &gMupScanner;
    Puts("<code>");
    parent->fn.reset();
    gScanner->translate(&(parent->fn));
    Puts("(");
    gScanner->translate(a[0]);
    Puts(")</code></li>\n");
    gScanner = sc;
    parent->inlist = false;
    return true;
  }
};

class MEngCmdMMethDet : public CmdHdl {
protected:
  MMethodDetailsEnv * parent;
public:
  MEngCmdMMethDet(MMethodDetailsEnv *p) : CmdHdl(0,1), parent(p) {}
  virtual bool translateArgs(Stream *s, SubStream *oa[],
			     int no, SubStream *a[]) {
    if(gLanguage != kEnglish)
      return true;
    if(!parent->inlist)
      Puts("<ul>\n");
    parent->inlist = true;
    localParaType _(kItem);
    putLangPara(a[0]);
    return true;
  }
};

class MDeCmdMMethDet : public CmdHdl {
protected:
  MMethodDetailsEnv * parent;
public:
  MDeCmdMMethDet(MMethodDetailsEnv *p) : CmdHdl(0,1), parent(p) {}
  virtual bool translateArgs(Stream *s, SubStream *oa[],
			     int no, SubStream *a[]) {
    if(gLanguage != kGerman)
      return true;
    if(!parent->inlist)
      Puts("<ul>\n");
    parent->inlist = true;
    localParaType _(kItem);
    putLangPara(a[0]);
    return true;
  }
};

static class MInfoEngCmd : public CmdHdl {
public:
  MInfoEngCmd() : CmdHdl(0,1) {}
  virtual bool translateArgs(Stream *s, SubStream *oa[],
			     int no, SubStream *a[]) {
    if(gLanguage != kEnglish)
      return true;
    gScanner->translate(a[0]);
    Puts("</h4>\n<ul>\n");
    return true;
  }
} minfoEngCmd;

static class MInfoDeCmd : public CmdHdl {
public:
  MInfoDeCmd() : CmdHdl(0,1) {}
  virtual bool translateArgs(Stream *s, SubStream *oa[],
			     int no, SubStream *a[]) {
    if(gLanguage != kGerman)
      return true;
    gScanner->translate(a[0]);
    Puts("</h4>\n<ul>\n");
    return true;
  }
} minfoDeCmd;

class MMethodDetailsOverloadsCmd : public CmdHdl {
protected:
  MMethodDetailsEnv * parent;
public:
  MMethodDetailsOverloadsCmd(MMethodDetailsEnv * p) :
    CmdHdl(1,0), parent(p) {}
  virtual bool translateArgs(Stream *s, SubStream *oa[],
			     int no, SubStream *a[]) {
    SubStream* pfn = &(parent->fn);
    SubStream** method = &pfn;
    if(!parent->inlist)
      Puts("<ul>\n");
    parent->inlist = true;
    parent->fn.reset();
    switch(gLanguage) {
    case kGerman:
      Puts("<li>Diese Methode &uuml;berl&auml;dt ");
      gCmds.top("Mup")->translateArgs(s, NULL, 0, no?oa:method);
      Puts(".</li>\n");
    case kEnglish:
      Puts("<li>This method overloads ");
      gCmds.top("Mup")->translateArgs(s, NULL, 0, no?oa:method);
      Puts(".</li>\n");
    }
    return true;
  }
};

bool
MMethodDetailsEnv::translateArgs(Stream *s, SubStream *oa[],
				 int no, SubStream *a[]) {
  Scanner *sc = gScanner;
  gScanner = &gMupScanner;
  Puts("\n<div class=\"MethodDetails\"><h4>");
  Puts(gDefs.top("method"));
  Puts(" ");
  gScanner->translate(a[0]);
  Puts(": ");
  gScanner = sc;

  fn = a[0];
  inlist = false;
  
  gCmds.push("infoEng", &minfoEngCmd);
  gCmds.push("infoDe",  &minfoDeCmd);

  gCmds.push("call", new McallCmdMMethDet(this));
  gCmds.push("eng", new MEngCmdMMethDet(this));
  gCmds.push("de", new MDeCmdMMethDet(this));
  gCmds.push("overloads", new MMethodDetailsOverloadsCmd(this));

  return true;
}

bool
MMethodDetailsEnv::translateEnd(Stream *s, SubStream *oa[], 
				int no, SubStream *a[]) {
  delete gCmds.pop("overloads");
  delete gCmds.pop("de");
  delete gCmds.pop("eng");
  delete gCmds.pop("call");

  gCmds.pop("infoDe");
  gCmds.pop("infoEng");
  if(inlist)
    Puts("</ul>\n");
  Puts("</ul></div>\n");
  return true;
}

// init the environment handlers

void
initMhelpEnvs() {
//    gEnvs.push("Mxcall", &mxcallEnv);
//    gEnvs.push("Mxresult", &mxresultEnv);
    gEnvs.push("Mexin", &mexinEnv);
    gEnvs.push("Mexout", &mexoutEnv);
    gEnvs.push("Mvin", &mvinEnv);
    //    gEnvs.push("Mvinn", &mvinnEnv);
    gEnvs.push("Mvout", &mvoutEnv);
    gEnvs.push("Mhtml", &mhtmlEnv);
    gEnvs.push("Mpackage", &mpackageEnv);
    gEnvs.push("Mhelppage", &mhelppageEnv);
    gEnvs.push("Mconstructor", &mhelppageEnv);
    gEnvs.push("Mseealso", &mseealsoEnv);
    gEnvs.push("Moverloading", &moverloadingEnv);
    gEnvs.push("Mgerman", &mgermanEnv);
    gEnvs.push("Menglish", &menglishEnv);
    gEnvs.push("Msyntax", &msyntaxEnv);
    gEnvs.push("MInfo", &mInfoEnv);
    gEnvs.push("MParameters", &mParametersEnv);
    gEnvs.push("MOptions", &mOptionsEnv);
    gEnvs.push("MDetails", &mDetailsEnv);
    gEnvs.push("MExample", &mExampleEnv);
    gEnvs.push("MOptionDetails", &mOptionDetailsEnv);
    gEnvs.push("MBackground", &mBackgroundEnv);
    gEnvs.push("MCalls", &mCallsEnv);
    gEnvs.push("MDomConst", &mDomConstEnv);
    gEnvs.push("MDomElConst", &mDomElConstEnv);
    gEnvs.push("MAxConst", &mAxConstEnv);
    gEnvs.push("MCatConst", &mCatConstEnv);
    gEnvs.push("MReturnValue", &mReturnValueEnv);
    gEnvs.push("MSideEffects", &mSideEffectsEnv);
    gEnvs.push("MFurtherDoc", &mFurtherDocEnv);
    gEnvs.push("MOperations", &mOperationsEnv);
    gEnvs.push("MEval", &mEvalEnv);
    gEnvs.push("MFuncCall", &mFuncCallEnv);
    gEnvs.push("MEnvSens", &mEnvSensEnv);
    gEnvs.push("MOperands", &mOperandsEnv);
    gEnvs.push("MOutput", &mOutputEnv);
    gEnvs.push("MChanges", &mchangesEnv);
    gEnvs.push("MMethodDetails", &mMethodDetailsEnv);
    gEnvs.push("MOverloading", &mOverloadingEnv);

    gSectCounter[1].addChild(&gHelppageCounter);
    gHelppageCounter.addChild(&gExampleCounter);
}

// end of file
