/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */


// Stream.cp -- character streams

#include <stdlib.h>
#include <string.h>
#include "error.h"
#include "stream.h"

Stream::Stream() { 
    lineNo = 1; fnam = "???";
}

Stream::~Stream() {}

void
Stream::fatal(char *s1, char *s2, char *s3) {
#ifdef MW
    fprintf(stderr, "Fatal error: %s%s%s [%s, line %ld]\nquit now\n", s1, s2, s3, fnam, lineNo);
#else
    fprintf(stderr, "Fatal error: %s%s%s [%s, line %ld]\n", s1, s2, s3, fnam, lineNo);
#endif
    doExit(1);
}

void
Stream::error(char *s1, char *s2, char *s3) {
    fprintf(stderr, "Error: %s%s%s [%s, line %ld]\n", s1, s2, s3, fnam, lineNo);
}

void
Stream::warning(char *s1, char *s2, char *s3) {
    fprintf(stderr, "Warning: %s%s%s [%s, line %ld]\n", s1, s2, s3, fnam, lineNo);
}

// FileStream represents an input file

FileStream::FileStream(FILE *f, char *n) : Stream() {
    inFile = f;
    fnam = n;
    startPos = ftell(inFile);
    actPos = startPos;
}

FileStream::~FileStream() {
    fclose(inFile);
}

char
FileStream::get() {
    int c = getc(inFile);
    if (c == EOF) return kEndChar;
    if (c == '\n') lineNo++;
    return (char) c;
}

StreamPos
FileStream::pos() {
    StreamPos p;
    p.pos.fpos = ftell(inFile);
    p.lineNo = lineNo;
    return p;
}

bool
FileStream::getRange(StreamPos p, std::string &str) {
  const int kBuflen=1024;
  static char buf[kBuflen];
  long act = ftell(inFile);
  // current character isn't read, it's the closing } or ]
  long diff = act - p.pos.fpos - 1;
  long read;
  
  str="";

  clearerr(inFile);
  fseek(inFile, p.pos.fpos, SEEK_SET);
  while(diff > 0 && !feof(inFile) && !ferror(inFile) ) {
    read = fread(buf, 1, 
		 (diff >= kBuflen) ? kBuflen-1 : diff,
		 inFile);
    buf[read]='\0';
    diff -= read;
    str += buf;
  }

  getc(inFile);

  return (0==diff);
}

bool
FileStream::getRange(StreamPos p, char *b, long n) {
    long act = ftell(inFile);
    fseek(inFile, p.pos.fpos, SEEK_SET);
    for (n--; ftell(inFile) != act; b++, n--) {
        if (n == 0) {
            *b = '\0';
            return false;
        }
        *b = (char) getc(inFile);
    }
    *b = '\0';
    return true;
}

void
FileStream::unget(char c) {
    if (c == '\n') lineNo--;
    ungetc((int) c, inFile);
}

bool
FileStream::end() {
    return (feof(inFile) ? true : false);
}

void
FileStream::reset() {
    fseek(inFile, startPos, SEEK_SET);
    lineNo = 1;
    actPos = startPos;
    actLine = lineNo;
}

void
FileStream::save() {
    actPos = ftell(inFile);
    actLine = lineNo;
}

void
FileStream::restore() {
    fseek(inFile, actPos, SEEK_SET);
    lineNo = actLine;
}

// SubStream represents a 'substring' of the input inFile

SubStream::SubStream() : Stream() {
    contPtr = strcpy(new char[1], "");
    actPos = contPtr;
    startLine = 1;
    lineNo = 1;
}

SubStream::SubStream(Stream *s, std::string &str) : Stream() {
  contPtr = strcpy(new char[str.length()+1], str.c_str());
  actPos = contPtr;
  startLine = s->line();
  lineNo = startLine;
  fnam = s->name();
}

SubStream::SubStream(Stream *s, char *c) : Stream() {
    contPtr = strcpy(new char[strlen(c)+1], c);
    actPos = contPtr;
    startLine = s->line();
    lineNo = startLine;
    fnam = s->name();
}

SubStream::SubStream(Stream *s, std::string &str, StreamPos *p) : Stream() {
  contPtr = strcpy(new char[str.length()+1], str.c_str());
  actPos = contPtr;
  startLine = p->lineNo;
  lineNo = startLine;
  fnam = s->name();
}


SubStream::SubStream(Stream *s, char *c, StreamPos *p) : Stream() {
    contPtr = strcpy(new char[strlen(c)+1], c);
    actPos = contPtr;
    startLine = p->lineNo;
    lineNo = startLine;
    fnam = s->name();
}

SubStream&
SubStream::operator=(SubStream *s) {
    delete contPtr;
    contPtr = strcpy(new char[strlen(s->cont())+1], s->cont());
    actPos = contPtr;
    startLine = s->line();
    lineNo = startLine;
    fnam = s->name();
    return *this;
}

SubStream&
SubStream::operator=(const char *c) {
    delete contPtr;
    contPtr = strcpy(new char[strlen(c)+1], c);
    reset();
    return *this;
}

char
SubStream::get() {
    char c = *actPos;
    if (c == '\0') return kEndChar;
    if (c == '\n') lineNo++;
    actPos++;
    return c;
}

StreamPos
SubStream::pos() {
    StreamPos p;
    p.pos.cpos = actPos;
    p.lineNo = lineNo;
    return p;
}

bool
SubStream::getRange(StreamPos p, std::string& str) {
  str = "";
  for( ; p.pos.cpos < actPos-1; p.pos.cpos++ )
    str += *p.pos.cpos;

  return true;
}

bool
SubStream::getRange(StreamPos p, char *b, long n) {
    for (n--; p.pos.cpos < actPos; p.pos.cpos++, b++, n--) {
        if (n == 0) {
            *b = '\0';
            return false;
        }
        *b = *p.pos.cpos;
    }
    *b = '\0';
    return true;
}

bool
SubStream::copy(char *b, int n) {
    strncpy(b, contPtr, n-1);
    b[n] = '\0';
    return ((strlen(contPtr) < (size_t)n) ? true : false);
}

void
SubStream::unget(char c) {
    actPos--;
    *actPos = c;
    if (c == '\n') lineNo--;
}

bool
SubStream::end() {
    return ((*actPos == '\0') ? true : false);
}

void
SubStream::reset() {
    actPos = contPtr;
    lineNo = startLine;
}

SubStream gNilStream;

// end of file
