/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */


// SymTab.h -- symbol table class

#ifndef _symtab_h_
#define _symtab_h_

#include "string.h"
#include "stack.h"
#include "list.h"

const int kBuckets = 256;

// for each symbol table entry there is a list of values

template<class T>
class SymStackLink : public StackLink {
    T *entryPtr;
public:
    SymStackLink() : StackLink() { entryPtr = NULL; }
    SymStackLink(T *e) : StackLink() { entryPtr = e; }
    T *entry() { return entryPtr; }
};

// a symbol table entry is a stack of values

template<class T>
class SymTabEntry : public Stack<SymStackLink<T> >, public ListLink {
    char *key;
public:
    SymTabEntry(const char *, T *);
    bool hasKey(const char *k) { return (strcmp(key, k) == 0); }
};

// finally the symbol table class

template<class T>
class SymTab {
    IList<SymTabEntry<T> > bucket[kBuckets];
    int hashVal(const char *);
    SymTabEntry<T> *findEntry(const char *);
public:
    void push(const char *, T *);
    T *pop(const char *);
    T *top(const char *);
};

// the methods 

template<class T>
SymTabEntry<T>::SymTabEntry(const char *k, T *v) : 
	Stack<SymStackLink<T> >(), ListLink() {
    key = strcpy(new char[strlen(k)+1], k);
    push( new SymStackLink<T>(v) );
}

template<class T>
int
SymTab<T>::hashVal(const char *k) {
    unsigned int i = 0;
    for (; *k != '\0'; k++) {
        i ^= (unsigned int) (*k);
    }
    return (i % kBuckets);
}

template<class T>
SymTabEntry<T> *
SymTab<T>::findEntry(const char *k) {
    IList<SymTabEntry<T> > *l = &bucket[hashVal(k)];
    SymTabEntry<T> *e;
    for (e = l->head(); e; e = l->next()) {
        if (e->hasKey(k)) return e;
    }
    return NULL;
}

template<class T>
void
SymTab<T>::push(const char *k, T *v) {
    SymTabEntry<T> *e = findEntry(k);
    if (e) {
        e->push( new SymStackLink<T>(v) );
    } else {
	e = new SymTabEntry<T>(k, v);
        bucket[hashVal(k)].insert( e );
    }
}

template<class T>
T *
SymTab<T>::top(const char *k) {
    SymTabEntry<T> *e = findEntry(k);
    if (e) {
        SymStackLink<T> *l = e->top();
        if (l) {
            return l->entry();
        }
    }
    return NULL;
}

template<class T>
T *
SymTab<T>::pop(const char *k) {
    SymTabEntry<T> *e = findEntry(k);
    if (e) {
        SymStackLink<T> *l = e->pop();
        T *v = l->entry();
        delete l;
        return v;
    }
    error("internal (SymTab::pop)");
    return NULL;
}

#endif // _symtab_h_
